@@ -32,8 +32,10 @@ my $configfile="/etc/uif/uif.conf";
 my @mapping = (	[ 'n', 'uifid', 'Name' ],
 		[ 's', 'uifsource', 'Source'],
 		[ 'i', 'uifindevice', 'InputInterface'],
+		[ 'pi', 'uifpindevice', 'PhysicalInputInterface'],
 		[ 'd', 'uifdest', 'Destination'],
 		[ 'o', 'uifoutdevice', 'OutputInterface'],
+		[ 'po', 'uifpoutdevice', 'PhysicalOutputInterface'],
 		[ 'p', 'uifservice', 'Service'],
 		[ 'm', 'uifmark', 'MarkMatch'],
 		[ 'S', 'uiftranssource', 'TranslatedSource'],
@@ -75,7 +77,7 @@ sub readConfig {
 				next;
 			}
 			chomp($line);
-			if ($line =~ /^([a-z0-9-]+)\s+(\d+)\s+/) {
+			if ($line =~ /^([a-z0-9-.]+)\s+(\d+)\s+/) {
 				$$Protocols{$1}=$2;
 				$$Protocols{$2}=$1;
 			} else {
@@ -149,7 +151,7 @@ sub readConfig {
 					my $entry;
 					foreach $entry (split(/\s+/, $parameter)) {
 						$entry eq '' && next;
-						if ($entry =~ /^([a-zA-Z])=([^=]+)$/) {
+						if ($entry =~ /^([a-zA-Z]{1,2})=([^=]+)$/) {
 							if (exists($charstringmap{$1})) {
 								my $value = $2;
 								$value =~ tr /,/ /;
@@ -241,7 +243,7 @@ sub expandRange {
 			my @temprange;
 			foreach $range (@rangearray) {
 				foreach (@{$rangehash{$range}}) {
-					my $last=$_+$range;
+					my $last=$_+$range-1;
 					push(@temprange, "$_:$last");
 				}
 			}
@@ -259,6 +261,7 @@ sub simplifyNetworks {
 	my $network;
 	if (@networks) {
 		my $ip;
+		my $onlymacs=1;
 		foreach (@networks) {
 			my $ip;
 			if ($_ =~ /(^[^=]+)=([^=]+)$/ ) {
@@ -269,21 +272,24 @@ sub simplifyNetworks {
 				push (@{$macs{$ip}}, $2);
 			} else {
 				$ip=new NetAddr::IP ($_) || die "not a valid network: $_\n";
+				$onlymacs=0;
 			}
 			push(@netobjects, $ip);
 		};
-		$netref = NetAddr::IP::compactref(\@netobjects);
+		if ($onlymacs==0) {
+			$netref = NetAddr::IP::compactref(\@netobjects);
 
-		@networks=();
-		foreach $network (@$netref) {
-			if (exists($macs{$network})) {
-				foreach $mac (@{$macs{$network}}) {
-					push (@networks, $network."=".$mac);
+			@networks=();
+			foreach $network (@$netref) {
+				if (exists($macs{$network})) {
+					foreach $mac (@{$macs{$network}}) {
+						push (@networks, $network."=".$mac);
+					}
+				} else {
+					push (@networks, $network);
 				}
-			} else {
-				push (@networks, $network);
-			}
-		};
+			};
+		}
 	}
 	return (@networks);
 }
@@ -548,11 +554,11 @@ sub validateData {
 			}
 			$$rule{'MarkMatch'}=\@array;
 		}
-		foreach $hashentry (qw(InputInterface OutputInterface)) {
+		foreach $hashentry (qw(InputInterface OutputInterface PhysicalInputInterface PhysicalOutputInterface)) {
 			if (exists($$rule{$hashentry})) {
-				if ($hashentry eq 'InputInterface' && $$rule{'Type'} =~ /(OUTPUT|POSTROUTING)/) {
+				if (($hashentry eq 'InputInterface' || $hashentry eq 'PhysicalInputInterface') && $$rule{'Type'} =~ /(OUTPUT|POSTROUTING)/) {
 					die "can't use input interface in output rule:\n$$rule{'Rule'}\n";
-				} elsif ($hashentry eq 'OutputInterface' && $$rule{'Type'} =~ /(INPUT|PREROUTING)/) {
+				} elsif (($hashentry eq 'OutputInterface' || $hashentry eq 'PhysicalOutputInterface') && $$rule{'Type'} =~ /(INPUT|PREROUTING)/) {
 					die "can't use output interface in input rule:\n$$rule{'Rule'}\n";
 				}
 				my $position;
@@ -876,6 +882,8 @@ sub genRuleDump {
 		my @destination;
 		my @inputinterface;
 		my @outputinterface;
+		my @physicalinputinterface;
+		my @physicaloutputinterface;
 		my @mark;
 		my $action;
 		my $logaction;
@@ -1092,6 +1100,28 @@ sub genRuleDump {
 				push (@outputinterface, "-o $not $output");
 			}
 		}
+		if (exists($$rule{'PhysicalInputInterface'})) {
+			if (exists($$rule{'PhysicalInputInterface-not'})) {
+				$not='!';
+			} else {
+				$not='';
+			}
+			my $input;
+			foreach $input (@{$$rule{'PhysicalInputInterface'}}) {
+				push (@physicalinputinterface, "-m physdev --physdev-in $not $input");
+			}
+		}
+		if (exists($$rule{'PhysicalOutputInterface'})) {
+			if (exists($$rule{'PhysicalOutputInterface-not'})) {
+				$not='!';
+			} else {
+				$not='';
+			}
+			my $output;
+			foreach $output (@{$$rule{'PhysicalOutputInterface'}}) {
+				push (@physicaloutputinterface, "-m physdev --physdev-out $not $output");
+			}
+		}
 		if (exists($$rule{'MarkMatch'})) {
 			my $mark;
 			foreach $mark (@{$$rule{'MarkMatch'}}) {
@@ -1128,7 +1158,7 @@ sub genRuleDump {
 		if (exists($$rule{'Limit'})) {
 			$action=" -m limit --limit $$rule{'Limit'} --limit-burst $$rule{'Limit-burst'} $action";
 		}
-		my @rulearray = (\@inputinterface, \@outputinterface, \@protocol, \@source, \@destination, \@mark);
+		my @rulearray = (\@inputinterface, \@outputinterface, \@physicalinputinterface, \@physicaloutputinterface, \@protocol, \@source, \@destination, \@mark);
 
 		my $level=1;
 		my $again=1;
